
# bf.sed - a simple brainf*ck interpreter in sed


# read the program and accumulate to hold buffer
# by convention, ! denotes the end of program and start of input.
# also, ? is a debug point

/!/!s/[^][<>.,+?!-]\{1,\}//g
H
$ b run
/!/ b run
d

: run
s/.*/#/
x
s/\n//g
/!/{
  # store any residual input in the hold buffer
  h
  s/!.*//
  s/[^][<>.,+?-]\{1,\}//g
  s/^/!/
  x
  s/[^!]*!//
  b tohex  
  : begin
  s/!//
}

# strip comments of the form [...] after a ]
s/]\[[^][]*]/]/g

# rewrite shortcuts
s/\[-]/z/g
s/\[<]/l/g
s/\[>]/r/g
s/\[<<]/L/g
s/\[>>]/R/g
#s/\[>>+<<-]/{\[>>+<<-]}/g

# try unrolling loops

:unroll
/\[\([-+<>]*\)]/{
  s//{@_\1}/
  :uskip
  s/\(@[<>]*_\)\([^<>}]\{1,\}\)/\2\1/
  s/\(@[<>]*\)_\(>\{1,\}\)/\2\1\2_/
  s/@\([<>]*_\)\(<\{1,\}\)/\2@\2\1/
  s/<>//g
  t uskip
  /@_}/{
    # this loop contains the same number of < and >.
    # unroll it:
    s/{\([^{}]*\)@_}/p@\1}/
    :convert
    s/@\([^}+-]\{1,\}\)/\1@/
    s/@+/a@/
    s/@-/s@/
    t convert
    s/@}/d/
    b unroll
  }
  # else, this loop cannot be unrolled. keep the {} and try another loop
  s/@[<>]*_//
  b unroll
}
# no candidate left, restore the []
y/{}/[]/

# initialize the data list.
s/$/#~00/

# fetch first instruction
s/./&_&/

:eval


# copy the program instruction to the beginning and advance one step
s/.\([^_]*\)_\(.\)/\2\1\2_/

# shortcuts
/^[zlrLR][^~]*~00/b eval

# z - zero [-]
/^z/{ 
  s/~../~00/
  b eval
}

# l - move left [<]
/^l/{
  /#0*~/ b error_left
  s/\(#.*\)00\(\(..\)*\)~/\1~00\2/
  b eval
}

# L - move doubly left [<<]
/^L/{
  /#.\{0,2\}~/ b error_left
  s/\(#.*\)00\(..\(....\)*\)~/\1~00\2/
  /#..~0\{0,1\}[^0]/ b error_left
  b eval
}

# r - move right [>]
/^r/{
  s/~\(\(\([^0].\)*\(0[^0]\)*\)*\)/\1~/
  s/~$/~00/
  b eval
}

# R - move doubly right [>>]
/^R/{
  s/~\(\(\([^0]...\)*\(0[^0]..\)*\)*\)/\1~/
  /~00/! s/~\(..\)$/\100~00/
  s/~$/~00/  
  b eval
}

# p - push number and start unrolled loop
/^p/{
  /~00/ {
    # skip this loop entirely
    s/_\([^d]*d\)/\1_/
    b eval
  }
  s/\(.\)\([^~]*~\)\(..\)/\1\3\2\3/
  b eval
}


# d - pop number and end loop
/^d/{
  s/\(.\)../\1/
  /~00/!s/\(p[^p_]*\)_/_\1/
  b eval
}

# a - add
/^a/{
  /~00/ {
    # simple case, x + 0 => x
    s/\(.\)\(..\)\([^~]*~\)00/\1\2\3\2/
    b eval
  }
  
  s/\(.\)\(..\)\([^~]*~\)\(..\)/0\2,0\4,&/

  # start of add. 01,23, => 34, at the beginning of the pattern space
  :add
  s/\([^,]*,[^,]*\),/\1:fedcba9876543210:fedcba9876543210:/
  :addloop
  # 1     2        3        4     5        6          7         8
  s/\(.\),\([^:]*\)\([^:]\):\(0*\)\([^:]*\1\([^:]*\)\)\(:[^:]*\3\([^:]*\)\)/\
\4\6\8\5\5,\2:\5\7/
  s/\n.\{31\}\(.\)[^,]\{16\}[^,]*,\([^:]*:\)/,\1\20/
  s/\(0*\)\n.\{31\}\(.\)[^,]*,\([^:]*\)\1:/,\1\2\3:/
  t addloop
  s/,\([^,]*\):[^:]*:[^:]*:/\1,/
  # end of add.
  
  
  /^[^0]/ b error_plus
  s/^0\(..\),\([^~]*~\)../\2\1/
  b eval
}

# s - sub
/^s/{
  /\(.\)\(..\)\([^~]*~\)\2/ {
    # simple case, x - x => 0
    s//\1\2\300/
    b eval
  }
  
  s/\(.\)\(..\)\([^~]*~\)\(..\)/,\2:fedcba9876543210,0123456789abcdef:,f\4,&/
  :not
  s/,\(.\)\([^:]*:[^:]*\1[^:]\{16\}\([^:]\)\)/\3,\2/
  /,:/!t not
  
  s/,:[^:]*:/!0123456789abcdef10ff00f0:/
  s/\(.\)\(f*\)![^f]*\1\(.0*\)[^:]*\2\(0*\)[^:]*:/\3\4/
  s/^/0/  
  
  b add
}



/^#/ b end

/^?/{
  s/[^#]*#\(.*\)/\1\
&/
  s/~/~ /
  s/../& /g
  s/~  /~/
  s/ *\(\n\)/\1/
  P
  s/.*\n//
  s/ //g
  b eval
}

/^\[/{
  /~00/{
    # skip to the matching ]
    s/_/_@/
    : skip
    s/\(_@@*\)\([^]@[]\{1,\}\)/\2\1/
    s/\(_@@*\)\(\[\)/\2\1@/
    s/\(_@*\)@\(]\)/\2\1/
    t skip
    /@/{
      : error_brackets
      s/.*/Error: mismatched brackets/
      q
    }
    b eval
  }
  # else drop a = mark after the [
  s/_/=_/
  b eval
}

/^]/{
  /=/! b error_brackets
  s/~$/~00/
  /~00/!{
    # skip to the matching [
    s/^\([^_]*=\)\([^_]*\)_/\1_\2/
    b eval
  }
  # else drop the = mark
  s/^\([^_]*\)=/\1/
  b eval
}

/^>/{
  s/~\(..\)/\1~/
  s/~$/~00/
  b eval
}

/^</{
  /#~/{
    :error_left
    s/.*/Error: pointer moved too far left/
    q
  }
  s/\(.[^#]\)~/~\1/
  b eval
}


/^-/{
  # decrement number
  /~00/{
    :error_minus
    s/.*/Error: number decremented below zero/
    q
  }
  s/~../&!fedcba98765432100ff0f:/
  s/\(.\)\(0*\)![^0]*\1\(.f*\)[^:]*\2\(f*\)[^:]*:/\3\4/

  b eval
}

/^+/{
  # increment number
  /~ff/{
    :error_plus
    s/.*/Error: number incremented past 255/
    q
  }
  s/~../&!0123456789abcdef10ff00f0:/
  s/\(.\)\(f*\)![^f]*\1\(.0*\)[^:]*\2\(0*\)[^:]*:/\3\4/
  b eval
}

/^\./{
  # append the value to the hold buffer
  H
  x
  s/\n[^~]*~\(..\).*/\1/
  /0a$/b flush
  x
  
  b eval
}

/^,/{
  # fetch a value from the hold buf.
  # hold buffer contains: remaining input (hex), #, pending output (hex)
  x
  /^#/{
    # no remaining input
    $ {
      # no effect if at EOF.
      x
      b eval
    }
    # is there pending output?
    /#./{
      
      : flush
      # flush pending output
      s/#/&_/
      
      # convert from hex to bytes

: loop
s/_5f/\
_/
s/_2c/,_/
s/_0a/_/
t loop
s/_./&,\
x\x01\x02\x03\x04\x05\x06\x07\x08\x09x\x0b\x0c\x0d\x0e\x0f0,\
\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f1,\
 !"#$%\&'()*+x-.\/2,0123456789:;<=>?3,@ABCDEFGHIJKLMNO4,PQRSTUVWXYZ[\\]^x5,\
`abcdefghijklmno6,pqrstuvwxyz{|}~7,8,9,\
a,b,c,d,\
e,f,_/
s/_\(.\)[^_]*,\n*\([^,]*\)\1,[^_]*_\(.\)/_\30123456789abcdef,\2_/
s/_\(.\)[^_]*\1.\{16\}\(.\)[^_]*_/\2_/
t loop
s/_$//
s/\n/_/g

      # where are we coming from?
      x
      /^#/{
        # end of program
        x
        s/[^#]*#//
        
        # attempt to quit without printing a line
        ${
          /^$/d
        }
        q
      }
      x
      
      s/\([^#]*#\)\(.*\)/\2\
\1/
      P
      s/.*\n//
      
      x
      /^\./{
        # printing
        b eval
      }
      x
    }
    
    # read next line
    N
    s/.*\n//
    
    # convert from bytes to hex:
    : tohex

s/_/\
/g
s/^/_/
: loop2
s/_\n/5f_/
s/_,/2c_/
t loop2
s/_./&\
x\x01\x02\x03\x04\x05\x06\x07\x08\x09x\x0b\x0c\x0d\x0e\x0f0,\
\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f1,\
 !"#$%\&'()*+x-.\/2,0123456789:;<=>?3,@ABCDEFGHIJKLMNO4,PQRSTUVWXYZ[\\]^x5,\
`abcdefghijklmno6,pqrstuvwxyz{|}~7,8,9,\
a,b,c,d,\
e,f,_/
s/_\(.\).*\1\([^,]*\)\([0-9a-f]\),[^_]*_/\3_\2,0123456789abcdef_/
s/_.\{16\}\(.\)[^_]*_/\1_/
t loop2
s/_$/0a/
    

    # append the #
    s/$/#/
  }
  x
  /^!/{
    x
    s/0a#/#/
    x
    b begin
  }
  
  G
  s/~..\(.*\)\n\(..\).*/~\2\1/
  x
  s/^..//
  x
  b eval
}

s/.*/Error: unexpected character/
q

:end
x
b flush
s/.*//
q

